using System;
using System.IO;
using System.Windows.Forms;
using System.Globalization;
using Zcu.Mve.Core;
using Zcu.Mve.Strings;

namespace Zcu.Mve.Visualization
{
    /// <summary>
    /// Tda pro naten bod z DEM souboru.
    /// </summary>
    class DEMLoader : Module
    {
        char[] buffer;
        /// <summary>
        /// Jmno vstupnho souboru.
        /// </summary>
        private string filename = null;
        /// <summary>
        /// Uruje zmnu vstupnho souboru(jeho nzvu).
        /// </summary>
        private bool newFile = false;
        /// <summary>
        /// Vstupn data.
        /// </summary>
        private UniformDataArray output = null;

        public string FileName {
            get { return filename; }
            set {
                if (!File.Exists(value))         
                {
                    MessageBox.Show("Invalid file name!\n\nClick Cancel or choose existing file.", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                this.filename = value;
                this.newFile = true;
            }
        }

        /// <summary>
        /// Konstruktor vytvejc porty.
        /// </summary>
        public DEMLoader()
        {
            this.AddInPort("filename", typeof(MveString), false);
            this.AddOutPort("Point3D array", typeof(UniformDataArray));
        }

        /// <summary>
        /// Vstupn bod modulu.
        /// </summary>
        public override void Execute(){
            if (this.GetInPort("filename").Link != null)
            {
                if (!IsAllDataSame())
                {
                    this.FileName = ((MveString)GetInput("filename")).Value;
                }
            }
            if (newFile)
            {
                newFile = false;
                readPoints();
                SetOutput("Point3D array", output);
            }
        }

        /// <summary>
        /// Nate body z DEM souboru.
        /// </summary>
        private void readPoints() {
            StreamReader reader = new StreamReader(FileName);
            buffer = new char[1024];
            reader.ReadBlock(buffer, 0, 1024);
            int groundUnit = readGroundUnit();
            int elevationUnit = readElevationUnit();
            double[] resolution = readResolution();
            int cols = readCols();
            reader.ReadBlock(buffer, 0, 1024);
            int rows = readRows();
            output = new UniformDataArray(typeof(Point3D), rows * cols);
            int j, pos, x, y, z;
            double[] firstPoint;
            for (int i = 0; i < cols; i++ )
            {
                j = 0;
                pos = 144;
                firstPoint = readFirstPoint();
                while(j != rows ){
                    z =(int) resolution[2] * readNextElevation(pos);
                    x =(int) firstPoint[0];
                    y =(int) (firstPoint[1] + j * resolution[1]); 
                    output[i * rows + j] = new Point3D(x, y, z);
                    pos += 6;
                    j++;
                    if(pos >= 1024 - 5){
                        reader.ReadBlock(buffer, 0, 1024);
                        pos = 0;
                    }
                }
                reader.ReadBlock(buffer, 0, 1024);
            }
            reader.Close();
        }

        /// <summary>
        /// Pete jednotky pouit pro souadnice X a Y.
        /// </summary>
        /// <returns>Kd danch jednotek.</returns>
        private int readGroundUnit() {
            return Int32.Parse(new String(buffer, 529, 6));
        }

        /// <summary>
        /// Pete jednotky pouit pro souadnici Z.
        /// </summary>
        /// <returns>Kd dan jednotky.</returns>
        private int readElevationUnit() {
            return Int32.Parse(new String(buffer, 535, 6));
        }

        /// <summary>
        /// Pete rozlien ve vech smrech.
        /// </summary>
        /// <returns>Pole odpovdajc danmu rozlien.</returns>
        private double[] readResolution() {
            double[] ret = new double[3];
            NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
            
            ret[0] = Double.Parse(new String(buffer, 816, 12), NumberStyles.Float, nfi);
            ret[1] = Double.Parse(new String(buffer, 828, 12), NumberStyles.Float, nfi);
            ret[2] = Double.Parse(new String(buffer, 840, 12), NumberStyles.Float, nfi);

            return ret;
        }

        /// <summary>
        /// Pete poet sloupc..
        /// </summary>
        /// <returns>Poet sloupc.</returns>
        private int readCols() {
            return Int32.Parse(new String(buffer, 858, 6));
        }

        /// <summary>
        /// Pete poet dk.
        /// </summary>
        /// <returns>Poet dk.</returns>
        private int readRows() {
            return Int32.Parse(new String(buffer, 12, 6));
        }

        /// <summary>
        /// Nate souadnici Z.
        /// </summary>
        /// <param name="pos">Pozice v souboru, kde dan souadnice zan.</param>
        /// <returns>Souadnice Z.</returns>
        private int readNextElevation(int pos) {
            return Int32.Parse(new String(buffer, pos, 6));
        }

        /// <summary>
        /// Pete souadnice X a Y prvnho bodu v danm sloupci.
        /// </summary>
        /// <returns>Souadnice X a Y.</returns>
        private double[] readFirstPoint() {
            double[] ret = new double[2];
            NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
            
            ret[0] = Double.Parse(new String(buffer, 24, 20), NumberStyles.Float, nfi) * Math.Pow(10, Int32.Parse(new String(buffer, 45, 3)));
            ret[1] = Double.Parse(new String(buffer, 48, 20), NumberStyles.Float, nfi) * Math.Pow(10, Int32.Parse(new String(buffer, 69, 3)));

            return ret;
        }
    }
}
